unit RuleEditorForm;
//=== File Prolog ============================================================
//	This code was developed by RiverSoftAVG.
//
//--- Notes ------------------------------------------------------------------
//
//--- Development History  ---------------------------------------------------
//
//	02/2001	T. Grubb
//		           - Fixed Salience Error
//      11/2000 T. Grubb     Changed seSalience (TSpinEdit) to meSalience
//                           (TMaskEdit) since C++Builder does not have a
//                           TSpinEdit
//	04/2000	T. Grubb     Initial version.
//
//      File Contents:
//           TRulesEditDlg      Dialog Form for editing one rule or multiple rules
//           TIEDragObject      Drag and drop object for the TRulesEditDlg
//
//
//--- Warning ----------------------------------------------------------------
//	This software is property of RiverSoftAVG. Unauthorized use or
//      duplication of this software is strictly prohibited. Authorized users
//      are subject to the following restrictions:
//	*	RiverSoftAVG is not responsible for
//		any consequence of the use of this software.
//	*	The origin of this software must not be misrepresented either by
//		explicit claim or by omission.
//	*	Altered versions of this software must be plainly marked as such.
//	*	This notice may not be removed or altered.
//
//       2000-2001, Thomas G. Grubb
//
//=== End File Prolog ========================================================

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, StdCtrls, Buttons, ActnList, 
  InferenceEngine, GHashTable, IERule, IEFacts, Mask, GEdit,
  Menus, InterfaceCollections, IEInterfaces;

{$ObjExportAll On}

type
  TIEDragType = (idtNone, idtSlot, idtRule, idtTemplate, idtFunction);
  TIEDragObject = class(TDragObject)
  { Purpose: To hold the rule or template currently being dragged.  This class is
    created in the StartDrag event }
  private
    { Private declarations }
    FIEObject: TObject;
    FDragType: TIEDragType;
    FIEUnknown: IUnknown;
    function GetAsRule: TRule;
    function GetAsTemplate: IFactTemplate;
    function GetAsFactSlot: IFactSlot;
    function GetAsFunction: IUserFunction;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AObject: TObject); overload; virtual;
    constructor Create(AInterface: IUnknown); overload;
  published
    { Published declarations }
    property IEObject: TObject read FIEObject;
    property IEUnknown: IUnknown read FIEUnknown;
    property DragType: TIEDragType read FDragType;
    property AsRule: TRule read GetAsRule;
    property AsFactSlot: IFactSlot read GetAsFactSlot;
    property AsTemplate: IFactTemplate read GetAsTemplate;
    property AsFunction: IUserFunction read GetAsFunction;
  end; { TIEDragObject }

  TRulesEditDlg = class(TForm)
    SaveDialog1: TSaveDialog;
    OpenDialog1: TOpenDialog;
    StatusBar1: TStatusBar;
    ActionList1: TActionList;
    Panel1: TPanel;
    gbRules: TGroupBox;
    tvRules: TTreeView;
    Splitter3: TSplitter;
    pnlLoadSave: TPanel;
    bbSave: TBitBtn;
    bbLoad: TBitBtn;
    actNewRule: TAction;
    actReplaceRule: TAction;
    actDeleteRule: TAction;
    actAddRule: TAction;
    PopupMenu1: TPopupMenu;
    ExpandAll1: TMenuItem;
    CollapseAll1: TMenuItem;
    actExpandAll: TAction;
    actCollapseAll: TAction;
    pcDisplay: TPageControl;
    tsFactTemplates: TTabSheet;
    tvFactTemplates: TTreeView;
    tsFunctions: TTabSheet;
    lbFunctions: TListBox;
    Panel3: TPanel;
    eSearch: TEdit;
    bbClear: TBitBtn;
    actClearRules: TAction;
    Splitter1: TSplitter;
    Panel4: TPanel;
    pnlRuleEditor: TPanel;
    lblName: TLabel;
    lblComment: TLabel;
    lblSalience: TLabel;
    cbVariableAssign: TCheckBox;
    pcRuleEditor: TPageControl;
    tsEditor: TTabSheet;
    Splitter2: TSplitter;
    gbIF: TGroupBox;
    lbIF: TListBox;
    gbTHEN: TGroupBox;
    lbTHEN: TListBox;
    tsFreeForm: TTabSheet;
    Splitter4: TSplitter;
    GroupBox1: TGroupBox;
    mmThen: TMemo;
    GroupBox2: TGroupBox;
    mmIF: TMemo;
    eEdit: TEdit;
    eName: TGEdit;
    eComment: TGEdit;
    btnClear: TButton;
    Bevel1: TBevel;
    bbOk: TBitBtn;
    bbCancel: TBitBtn;
    bbHelp: TBitBtn;
    pnlTransfer: TPanel;
    sbNew: TSpeedButton;
    sbAdd: TSpeedButton;
    sbReplace: TSpeedButton;
    sbDelete: TSpeedButton;
    mmFunctionDesc: TMemo;
    Splitter5: TSplitter;
    meSalience: TMaskEdit;
    procedure FormCreate(Sender: TObject);
    procedure actNewRuleExecute(Sender: TObject);
    procedure lbIFDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure lbIFDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure tvFactTemplatesStartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure lbTHENDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure lbIFClick(Sender: TObject);
    procedure eEditChange(Sender: TObject);
    procedure eEditKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormDestroy(Sender: TObject);
    procedure bbSaveClick(Sender: TObject);
    procedure bbLoadClick(Sender: TObject);
    procedure actAddRuleExecute(Sender: TObject);
    procedure actReplaceRuleExecute(Sender: TObject);
    procedure actDeleteRuleExecute(Sender: TObject);
    procedure lbIFEnter(Sender: TObject);
    procedure lbIFKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btnClearClick(Sender: TObject);
    procedure actExpandAllExecute(Sender: TObject);
    procedure actCollapseAllExecute(Sender: TObject);
    procedure tvRulesStartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure lbFunctionsStartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure pcRuleEditorChange(Sender: TObject);
    procedure eEditDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure eSearchChange(Sender: TObject);
    procedure pcDisplayChange(Sender: TObject);
    procedure eSearchKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure lbFunctionsDblClick(Sender: TObject);
    procedure tvRulesChange(Sender: TObject; Node: TTreeNode);
    procedure FormShow(Sender: TObject);
    procedure actClearRulesExecute(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure tvRulesKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure lbFunctionsClick(Sender: TObject);
  private
    { Private declarations }
    FOldOnHint: TNotifyEvent;
    FIE: TInferenceEngine;
    FCurrentRule: TRule;
    FEditListBox: TListBox;
    FRuleVariables: TStringList;
    FMultiEdit: Boolean;
    function GetUserFunctions: TUserFunctions;
    procedure SetUserFunctions(const Value: TUserFunctions);
    procedure SetMultiEdit(const Value: Boolean);
    procedure OnHint(Sender: TObject);
    function GetFactTemplates: TFactTemplates;
    procedure SetFactTemplates(const Value: TFactTemplates);
    function GetRules: TRules;
    procedure SetRules(const Value: TRules);
    procedure SetCurrentRule(const Value: TRule);
    procedure UpdateRuleEditor;
    procedure SetEditListBox(const Value: TListBox);
    procedure OnAddRule( Sender: TObject; Rule: TRule );
    procedure OnRemoveRule( Sender: TObject; Rule: TRule );
    procedure OnNewFactTemplate(Sender: TObject; Template: IFactTemplate);
    function FindRule(Rule: TRule): TTreeNode;
    function GenerateUniqueBinding(AName: String): String;
    function GenerateUniqueVariable(DragObject: TIEDragObject): String;
    procedure UpdateCurrentRule;
    property EditListBox: TListBox read FEditListBox write SetEditListBox;
    property OldOnHint: TNotifyEvent read FOldOnHint write FOldOnHint;
    property RuleVariables: TStringList read FRuleVariables;
    property IE: TInferenceEngine read FIE;
  public
    { Public declarations }
    property CurrentRule: TRule read FCurrentRule write SetCurrentRule;
    property UserFunctions: TUserFunctions read GetUserFunctions write SetUserFunctions;
    property FactTemplates: TFactTemplates read GetFactTemplates write SetFactTemplates;
    property Rules: TRules read GetRules write SetRules;
    property MultiEdit: Boolean read FMultiEdit write SetMultiEdit default True;
  end;

var
  RulesEditDlg: TRulesEditDlg;

implementation

{$R *.DFM}

uses
    IEConsts;

var
   FirstShow: Boolean;

{ TIEDragObject }

constructor TIEDragObject.Create(AObject: TObject);
begin
     FIEObject := AObject;
     if IEObject is TRule then
          FDragType := idtRule
end;

constructor TIEDragObject.Create(AInterface: IUnknown);
var
   ATemp: IFactTemplate;
   ASlot: IFactSlot;
begin
     FIEUnknown := AInterface;
     if AInterface.QueryInterface( IFactTemplate, ATemp ) <> E_NOINTERFACE then
        FDragType := idtTemplate
     else if AInterface.QueryInterface( IFactSlot, ASlot ) <> E_NOINTERFACE then
        FDragType := idtSlot
     else
         FDragType := idtFunction;
end;

function TIEDragObject.GetAsFactSlot: IFactSlot;
begin
     result := FIEUnknown as IFactSlot;
end;

function TIEDragObject.GetAsFunction: IUserFunction;
begin
     result := IEUnknown as IUserFunction;
end;

function TIEDragObject.GetAsRule: TRule;
begin
     result := IEObject as TRule;
end;

function TIEDragObject.GetAsTemplate: IFactTemplate;
begin
     result := FIEUnknown as IFactTemplate;
end;

{ TRulesEditDlg }

function TRulesEditDlg.GetFactTemplates: TFactTemplates;
begin
     result := IE.FactTemplates;
end;

function TRulesEditDlg.GetRules: TRules;
begin
     result := IE.Rules;
end;

procedure TRulesEditDlg.SetFactTemplates(const Value: TFactTemplates);
begin
     IE.FactTemplates := Value;
     tvFactTemplates.Assign( IE.FactTemplates );
end;

procedure TRulesEditDlg.SetRules(const Value: TRules);
begin
     IE.Rules := Value;
end;

procedure TRulesEditDlg.FormCreate(Sender: TObject);
begin
     FIE := TInferenceEngine.Create( Self );
     FIE.OnAddRule := OnAddRule;
     FIE.OnRemoveRule := OnRemoveRule;
     FIE.OnNewFactTemplate := OnNewFactTemplate;
     // Save the OnHint and then assign it to our own event
     OldOnHint := Application.OnHint;
     Application.OnHint := OnHint;
     FMultiEdit := True;
     // Initialize the RuleVariables, this is used for guaranteeing unique names
     FRuleVariables := TStringList.Create;
     lbFunctions.Items.Assign( UserFunctions );
     FirstShow := True;
end;

procedure TRulesEditDlg.OnHint(Sender: TObject);
begin
     StatusBar1.Panels[0].Text := GetLongHint(Application.Hint);
end;

procedure TRulesEditDlg.actNewRuleExecute(Sender: TObject);
var
   ARule: TRule;
begin
     // Create a new rule and enable controls
     ARule := TRule.Create(IE);
     try
        CurrentRule := ARule;
     finally
        ARule.Free;
     end;
     ActiveControl := eName;
end;

procedure TRulesEditDlg.SetCurrentRule(const Value: TRule);
begin
     if CurrentRule <> Value then
     begin
          if CurrentRule <> nil then
          begin
               CurrentRule.Free;
               FCurrentRule := nil;
          end;
          if Value <> nil then
          begin
               FCurrentRule := TRule.Create( IE );
               CurrentRule.Assign( Value );
          end
          else
          begin
               FCurrentRule := nil;
          end;
          UpdateRuleEditor;
     end;
end;

procedure TRulesEditDlg.UpdateRuleEditor;
var
   i: Integer;
begin
     if CurrentRule = nil then
     begin
          // No rule, disable controls
          actAddRule.Enabled := False;
          actReplaceRule.Enabled := False;
          pnlRuleEditor.Visible := False;
     end
     else
     begin
          eEdit.Visible := False;
          RuleVariables.Clear;
          actAddRule.Enabled := True;
          actReplaceRule.Enabled := tvRules.Selected <> nil;
          pnlRuleEditor.Visible := True;
          eName.Text := CurrentRule.RuleName;
          eComment.Text := CurrentRule.Comment;
          meSalience.Text := IntToStr(CurrentRule.Salience);
          // IF Part
          lbIF.Items.Clear;
          mmIF.Lines.Clear;
          if CurrentRule.Patterns.Count > 0 then
             for i := 0 to CurrentRule.Patterns.Count - 1 do
                 mmIF.Lines.Add( lbIF.Items[lbIF.Items.Add( CurrentRule.Patterns[i].ToString )] );
          lbIF.Items.Add( '' );
          // THEN Part
          lbTHEN.Items.Clear;
          mmTHEN.Lines.Clear;
          if CurrentRule.Actions.Count > 0 then
             for i := 0 to CurrentRule.Actions.Count - 1 do
                 mmTHEN.Lines.Add( lbTHEN.Items[lbTHEN.Items.Add( CurrentRule.Actions[i].ToString )] );
          lbTHEN.Items.Add( '' );
          RuleVariables.Clear;
          // Add any bindings used by the rule to rule variables
          for i := 0 to CurrentRule.BindingCount - 1 do
              RuleVariables.Add( CurrentRule.Binding[i].Name );
     end;
end;

procedure TRulesEditDlg.lbIFDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
     Accept := IsDragObject(Source) or (Source = Sender) or (Source = tvRules)
               or (Source is TListBox) or (Source is TMemo);
end;

procedure TRulesEditDlg.lbIFDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
   Index: Integer;
   ListBox: TListBox;
   Memo: TMemo;
   AString: String;
   ASlot: IFactSlot;
begin
    // Find out where the cursor is located
    if Sender is TListBox then
    begin
         Index := TListBox(Sender).ItemAtPos( Point(x,y), False );
         if Index >= TListBox(Sender).Items.Count then
            Index := TListBox(Sender).Items.Count-1;
    end
    else //if Sender is TMemo then
        Index := 0;
    AString := '';
    // Find out which listbox and memo we are modifying
    if (Sender = lbIF) or (Sender = mmIF) then
    begin
         ListBox := lbIF;
         Memo := mmIF;
         // IF section, should we assign it to a variable
         if cbVariableAssign.Checked and (Source is TIEDragObject) and
            ((Source as TIEDragObject).DragType in [idtSlot, idtTemplate]) then
            AString := '?' + GenerateUniqueVariable( Source as TIEDragObject )+ ' <- ';
    end
    else //if (Sender = lbTHEN) or (Sender = mmTHEN) then
    begin
         ListBox := lbTHEN;
         Memo := mmTHEN;
    end;

    // If we are just dragging something within the object, just move it
    if Sender <> Source then
    begin
      if Source is TIEDragObject then
      with Source as TIEDragObject do
           case DragType of
             idtSlot:
             begin
                  AString := AString + '(' + AsFactSlot.GetTemplateName + ' ';
                  AString := AString + '(' + AsFactSlot.Name + ' ';
                  if cbVariableAssign.Checked and
                     ((Sender = lbIF) or (Sender = mmIF)) then
                     AString := AString + '?' + GenerateUniqueBinding( AsFactSlot.Name ) + '))'
                  else
                      AString := AString + AsFactSlot.SlotDefault.ToString + '))';
                  Memo.SelText := AString;
                  ListBox.Items.Insert( Index, AString );
             end;
             idtTemplate:
             begin
                  if not AsTemplate.Ordered then
                  begin
                       AString := AString + '(' + AsTemplate.TemplateName + ' ';
                       with AsTemplate.SlotNames do
                       while HasMoreElements do
                       begin
                            ASlot := AsTemplate.Slot[NextElement];
                            AString := AString + '(' + ASlot.Name + ' ';
                            if cbVariableAssign.Checked and
                               ((Sender = lbIF) or (Sender = mmIF)) then
                               AString := AString + '?' + GenerateUniqueBinding( ASlot.Name ) + ')'
                            else
                               AString := AString + ASlot.SlotDefault.ToString + ')';
                       end;
                       AString := AString + ')';
                  end
                  else
                       AString := AString + '(' + AsTemplate.TemplateName + ')';
                  Memo.SelText := AString;
                  ListBox.Items.Insert( Index, AString );
             end;
             idtRule:
             begin
                  Memo.SelText := AString + AsRule.ToString;
                  ListBox.Items.Insert( Index, AString + AsRule.ToString );
             end;
             idtFunction:
             begin
                  AString := '('+AsFunction.FunctionName+')';
                  Memo.SelText := AString;
                  ListBox.Items.Insert( Index, AString );
             end;
           end
      else if Source is TTreeView then
      begin
           Memo.SelText := AString + TTreeView(Source).Selected.Text;
           ListBox.Items.Insert( Index, AString + TTreeView(Source).Selected.Text );
      end
      else if Source is TListBox then
      begin
           TListBox(Sender).Items.Insert( Index, TListBox(Source).Items[TListBox(Source).ItemIndex] );
           TListBox(Source).Items.Delete( TListBox(Source).ItemIndex );
      end;
    end
    else // we are just dragging something within the object, just move it
    begin
         if Sender is TListBox then
            TListBox(Sender).Items.Move(TListBox(Sender).ItemIndex, Index)
         else if Sender is TMemo then
              ;
    end;
    ActiveControl := Sender as TWinControl;
end;

procedure TRulesEditDlg.tvFactTemplatesStartDrag(Sender: TObject;
  var DragObject: TDragObject);
var
   TreeNode: TTreeNode;
begin
     with Sender as TTreeView do
     begin
          TreeNode := Selected;
          if TreeNode = nil then Exit;
          if FactTemplates[String(TreeNode.Data)].Ordered or
             (TreeNode.Level = 0) then
             DragObject := TIEDragObject.Create( FactTemplates[String(TreeNode.Data)] )
          else
          begin
               while TreeNode.Level <> 1 do TreeNode := TreeNode.Parent;
               if not FactTemplates[String(TreeNode.Data)].ContainsSlot(TreeNode.Text) then
                  DragObject := TIEDragObject.Create( FactTemplates[String(TreeNode.Data)] )
               else
                   DragObject := TIEDragObject.Create( FactTemplates[String(TreeNode.Data)].Slot[TreeNode.Text] )
          end;
     end;
end;

procedure TRulesEditDlg.lbTHENDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
     Accept := ( IsDragObject(Source) and
                 ((Source as TIEDragObject).DragType in [idtSlot, idtTemplate, idtFunction])) or
               (Source is TTreeView) or
               (Source = Sender)
               or (Source is TListBox) or (Source is TMemo);  
end;

procedure TRulesEditDlg.lbIFClick(Sender: TObject);
begin
     with Sender as TListBox do
     begin
          if ItemIndex = -1 then
          begin
               eEdit.Visible := False;
               Exit;
          end;
          EditListBox := Sender as TListBox;
          eEdit.Text := Items[ItemIndex];
     end;
end;

procedure TRulesEditDlg.SetEditListBox(const Value: TListBox);
begin
     FEditListBox := Value;
     eEdit.Enabled := EditListBox <> nil;
     eEdit.Visible := EditListBox <> nil;
end;

procedure TRulesEditDlg.eEditChange(Sender: TObject);
begin
     with Sender as TEdit do
     begin
          if not Enabled then Exit;
          ActiveControl := Sender as TEdit;
          EditListBox.Items[EditListBox.ItemIndex] := Text;
     end;
end;

procedure TRulesEditDlg.eEditKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
     case Key of
          VK_UP:       if EditListBox.ItemIndex > 0 then
                       begin
                            EditListBox.ItemIndex := EditListBox.ItemIndex - 1;
                            EditListBox.OnClick(EditListBox);
                       end;
          VK_DOWN:     if EditListBox.ItemIndex < EditListBox.Items.Count - 1 then
                       begin
                            EditListBox.ItemIndex := EditListBox.ItemIndex + 1;
                            EditListBox.OnClick(EditListBox);
                       end;
          vk_Tab:      if EditListBox = lbIF then
                           ActiveControl := lbTHEN
                       else
                           ActiveControl := bbOK;
     end;
end;

procedure TRulesEditDlg.OnAddRule(Sender: TObject; Rule: TRule);
var
   i: Integer;
   Str: TStrings;
begin
     Str := TStringList.Create;
     try
        Str.Text := Rule.ToString;
        if Rule.Comment = '' then
           tvRules.Selected := tvRules.Items.AddObject( nil, Rule.RuleName, Rule )
        else
            tvRules.Selected := tvRules.Items.AddObject( nil, Rule.RuleName+'  "'+Rule.Comment+'"', Rule );
        for i := 2 to Str.Count - 1 do
            tvRules.Items.AddChildObject( tvRules.Selected, Trim(Str[i]), Rule );
     finally
        Str.Free;
     end;
end;

procedure TRulesEditDlg.OnRemoveRule(Sender: TObject; Rule: TRule);
var
  CurItem: TTreeNode;
begin
     if tvRules = nil then Exit;
     with tvRules do
     begin
          CurItem := FindRule( Rule );
          if CurItem <> nil then
             Items.Delete( CurItem );
          actClearRules.Enabled := IE.Rules.Count > 0;
     end;
end;

function TRulesEditDlg.FindRule( Rule: TRule ): TTreeNode;
begin
     result := nil;
     if tvRules = nil then Exit;
     with tvRules do
     begin
          result := Items.GetFirstNode;
          while (result <> nil) and (result.Data <> Rule) do
                result := result.GetNext;
     end;
end;

procedure TRulesEditDlg.UpdateCurrentRule;
var
   RuleString: String;
   ARule: TRule;
   Salience: Integer;
begin
     // Defrule name
     RuleString := '( defrule '+ eName.Text + NewLineString;
     // Comment
     if eComment.Text <> '' then
        RuleString := RuleString + '"' + eComment.Text + '"' + NewLineString;
     // Salience
     if StrToInt(Trim(meSalience.Text)) <> 0 then
        RuleString := RuleString + '(declare (salience ' + Trim(meSalience.Text) + '))' + NewLineString;
     case pcRuleEditor.ActivePage.TabIndex of
     0:
       begin
            // Patterns
            RuleString := RuleString + lbIF.Items.Text;
            // Then
            RuleString := RuleString + '=>' + NewLineString;
            // Actions
            RuleString := RuleString + lbTHEN.Items.Text + ')' + NewLineString;
       end;
     1:
       begin
            // Patterns
            RuleString := RuleString + mmIF.Lines.Text;
            // Then
            RuleString := RuleString + '=>' + NewLineString;
            // Actions
            RuleString := RuleString + mmTHEN.Lines.Text + ')' + NewLineString;
       end;
     end;

     // Ok, we have the string.  Lets try to parse and add it to the IE
     ARule := IE.ParseRule( RuleString );
     try
        CurrentRule := ARule;
     finally
        ARule.Free;
     end;
     //IE.AddRule( CurrentRule );
end;

procedure TRulesEditDlg.FormDestroy(Sender: TObject);
begin
     Application.OnHint := OldOnHint;
     CurrentRule := nil;
     RuleVariables.Free;
end;

procedure TRulesEditDlg.bbSaveClick(Sender: TObject);
{ Save the rules to a file }
begin
     // Get the filename from the user
     if SaveDialog1.Execute then
     begin
          // Save the file
          IE.SaveRules( SaveDialog1.Filename );
          OpenDialog1.FileName := SaveDialog1.Filename;
     end;
end;

procedure TRulesEditDlg.bbLoadClick(Sender: TObject);
{ load the rules from a file }
begin
     // Get the filename from the user
     if OpenDialog1.Execute then
     begin
          // Save the file
          IE.LoadRules( OpenDialog1.Filename );
          SaveDialog1.FileName := OpenDialog1.Filename;
     end;
end;

procedure TRulesEditDlg.actAddRuleExecute(Sender: TObject);
var
   i: Integer;
   ARule: TRule;
begin
     if EditListBox <> nil then i := EditListBox.ItemIndex
                           else i := -1;
     UpdateCurrentRule;
     // Add Rule or a copy?
     if IE.Rules.ContainsKey( CurrentRule.RuleName ) then
        raise EIEException.CreateFmt( SDuplicateRuleNames, [CurrentRule.RuleName] );
     ARule := TRule.Create(IE);
     ARule.Assign( CurrentRule );
     IE.Rules.Add( ARule );
     // Make the treeview select this rule
     tvRules.Selected := FindRule( ARule );
     CurrentRule := TRule(tvRules.Selected.Data);
     if EditListBox <> nil then EditListBox.ItemIndex := i;
end;

procedure TRulesEditDlg.actReplaceRuleExecute(Sender: TObject);
var
   CurrentString: String;
   ARule: TRule;
begin
     // Replace Selected rule with Current Rule
     if tvRules.Selected = nil then Exit;
     // First parse the new rule to make sure that we can (i.e., no syntax errors)
     UpdateCurrentRule;
     CurrentString := CurrentRule.ToString;
     actDeleteRuleExecute( Sender );
     ARule := IE.ParseRule( CurrentString );
     try
        CurrentRule := ARule;
     finally
        ARule.Free;
     end;
     actAddRuleExecute( Sender );
end;

procedure TRulesEditDlg.actDeleteRuleExecute(Sender: TObject);
begin
     if tvRules.Selected = nil then Exit;
     IE.Rules.Remove( TRule(tvRules.Selected.Data).RuleName );
end;

procedure TRulesEditDlg.lbIFEnter(Sender: TObject);
begin
     lbIFClick( Sender );
end;

function TRulesEditDlg.GenerateUniqueVariable(DragObject: TIEDragObject): String;
var
   Index: Integer;
begin
     result := '';
     case DragObject.DragType of
       idtSlot:
            result := DragObject.AsFactSlot.GetTemplateName +
                      DragObject.AsFactSlot.Name;
       idtTemplate:
            result := DragObject.AsTemplate.TemplateName;
     end;
     Index := 1;
     while RuleVariables.IndexOf( result+IntToStr(Index) ) > -1 do
           Inc(Index);
     result := result + IntToStr(Index);
     RuleVariables.Add( result );
end;

function TRulesEditDlg.GenerateUniqueBinding(AName: String): String;
var
   Index: Integer;
begin
     result := AName;
     Index := 1;
     while RuleVariables.IndexOf( result+IntToStr(Index) ) > -1 do
           Inc(Index);
     result := result + IntToStr(Index);
     RuleVariables.Add( result );
end;

procedure TRulesEditDlg.lbIFKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
     if Key in [vk_Delete, vk_Back] then
        with Sender as TListBox do
        begin
             // Delete the line if not end of listbox and blank
             if (ItemIndex = (Items.Count - 1)) and (Trim(Items[ItemIndex]) = '') then Exit;
             Items.Delete(ItemIndex);
             OnClick(Sender);
        end
     else
         ActiveControl := eEdit;
end;

procedure TRulesEditDlg.btnClearClick(Sender: TObject);
var
   ARule: TRule;
begin
     // Clear the rule
     ARule := TRule.Create(IE);
     try
        CurrentRule := ARule;
     finally
        ARule.Free;
     end;
end;

procedure TRulesEditDlg.actExpandAllExecute(Sender: TObject);
begin
     with PopupMenu1 do
          (PopUpComponent as TTreeView).FullExpand;
end;

procedure TRulesEditDlg.actCollapseAllExecute(Sender: TObject);
begin
     with PopupMenu1 do
          (PopUpComponent as TTreeView).FullCollapse;
end;

procedure TRulesEditDlg.tvRulesStartDrag(Sender: TObject;
  var DragObject: TDragObject);
begin
     with Sender as TTreeView do
     begin
          if Selected = nil then Exit;
          if (Selected.Level = 0) or (Selected.Parent.GetFirstChild = Selected) then
             DragObject := TIEDragObject.Create( TObject(Selected.Data) );
     end;
end;

procedure TRulesEditDlg.OnNewFactTemplate(Sender: TObject;
  Template: IFactTemplate);
begin
     // New template created, add it to the treeview and sort
     tvFactTemplates.Assign( IE.FactTemplates );
end;

function TRulesEditDlg.GetUserFunctions: TUserFunctions;
begin
     result := IE.UserFunctions;
end;

procedure TRulesEditDlg.SetUserFunctions(const Value: TUserFunctions);
begin
     IE.UserFunctions := Value;
end;

procedure TRulesEditDlg.lbFunctionsStartDrag(Sender: TObject;
  var DragObject: TDragObject);
begin
     with Sender as TListBox do
     begin
          if ItemIndex = -1 then Exit;
          if IE.UserFunctions[Items[ItemIndex]] = nil then
             Raise Exception.Create('Trouble finding '+Items[ItemIndex] );
          DragObject := TIEDragObject.Create( IE.UserFunctions[Items[ItemIndex]] );
     end;
end;

procedure TRulesEditDlg.pcRuleEditorChange(Sender: TObject);
begin
     eEdit.Visible := False;
     with Sender as TPageControl do
     case ActivePage.TabIndex of
     0:
       begin
            lbIF.Items := mmIF.Lines;
            lbTHEN.Items := mmTHEN.Lines;
            ActiveControl := lbIF;
       end;
     1:
       begin
            mmIF.Lines := lbIF.Items;
            mmTHEN.Lines := lbTHEN.Items;
            ActiveControl := mmIF;
       end;
     end;
end;

procedure TRulesEditDlg.eEditDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
   AString: String;
   ASlot: IFactSlot;
   i: Integer;
begin
    // If we are just dragging something within the object, just move it
    with Sender as TEdit do
    if Sender <> Source then
    begin
      if Source is TIEDragObject then
      with Source as TIEDragObject do
           case DragType of
             idtSlot:
             begin
                  // First check if edit box already has template, if so, then
                  // just add the slot
                  i := Pos(AsFactSlot.GetTemplateName,TEdit(Sender).Text);
                  if (i = 0) or
                     ((i >= TEdit(Sender).SelStart) and
                      (i <= (TEdit(Sender).SelStart+TEdit(Sender).SelLength))) then
                     AString := '(' + AsFactSlot.GetTemplateName + ' '
                  else
                      AString := '';
                  AString := AString + '(' + AsFactSlot.Name + ' ';
                  if cbVariableAssign.Checked and
                     ((Sender = lbIF) or (Sender = mmIF) or
                      (Sender = eEdit)) then
                     AString := AString + '?' + GenerateUniqueBinding( AsFactSlot.Name ) + '))'
                  else
                      AString := AString + AsFactSlot.SlotDefault.ToString + '))';
             end;
             idtTemplate:
             begin
                  if not AsTemplate.Ordered then
                  begin
                       AString := '(' + AsTemplate.TemplateName + ' ';
                       with AsTemplate.SlotNames do
                       while HasMoreElements do
                       begin
                            ASlot := AsTemplate.Slot[NextElement];
                            AString := AString + '(' + ASlot.Name + ' ';
                            if cbVariableAssign.Checked and
                               ((Sender = lbIF) or (Sender = mmIF) or
                                (Sender = eEdit)) then
                               AString := AString + '?' + GenerateUniqueBinding( ASlot.Name ) + ')'
                            else
                               AString := AString + ASlot.SlotDefault.ToString + ')';
                       end;
                       AString := AString + ')';
                  end
                  else
                       AString := '(' + AsTemplate.TemplateName + ')';
             end;
             idtRule:
             begin
                  AString := AsRule.ToString;
             end;
             idtFunction:
             begin
                  AString := '('+AsFunction.FunctionName+')';
             end;
           end
      else if Source is TTreeView then
      begin
           AString := TTreeView(Source).Selected.Text;
      end;
      i := SelStart + Length(AString);
      SelText := AString;
      SelStart := i-1;
    end;
end;

procedure TRulesEditDlg.eSearchChange(Sender: TObject);
var
   i: Integer;
begin
     with lbFunctions do
     begin
          if Items.Count = 0 then Exit;
          i := 0;
          while (i < Items.Count) and (TEdit(Sender).Text > Items[i]) do
                i := i + 1;
          ItemIndex := i;
          mmFunctionDesc.Lines.Text := IE.UserFunctions[Items[ItemIndex]].Comment;
     end;
end;

procedure TRulesEditDlg.pcDisplayChange(Sender: TObject);
begin
     with Sender as TPageControl do
          case ActivePage.TabIndex of
          0: ActiveControl := tvFactTemplates;
          1: ActiveControl := eSearch;
          end;
end;

procedure TRulesEditDlg.eSearchKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
     if Key = vk_Return then
     begin
          (Sender as TEdit).Text := lbFunctions.Items[lbFunctions.ItemIndex];
          TEdit(Sender).SelStart := Length(lbFunctions.Items[lbFunctions.ItemIndex]);
          lbFunctionsDblClick( lbFunctions );
          Key := 0;
     end;
end;

procedure TRulesEditDlg.lbFunctionsDblClick(Sender: TObject);
var
   DragObject: TIEDragObject;
begin
     with Sender as TListBox do
     begin
          if (ItemIndex = -1) or (not eEdit.Visible) then Exit;
          DragObject := TIEDragObject.Create( IE.UserFunctions[Items[ItemIndex]] );
          try
             eEditDragDrop( eEdit, DragObject, 0, 0 );
          finally
             DragObject.Free;
          end;
     end;
end;

procedure TRulesEditDlg.tvRulesChange(Sender: TObject; Node: TTreeNode);
begin
     // Change the current rule to the selected rule
     with Sender as TTreeView do
     begin
          actClearRules.Enabled := IE.Rules.Count > 0;
          actDeleteRule.Enabled := Selected <> nil;
          actReplaceRule.Enabled := (Selected <> nil) and (CurrentRule <> nil);
          if Selected = nil then Exit;
          if Selected.Level > 0 then Exit;
          CurrentRule := TRule(Selected.Data);
          actReplaceRule.Enabled := (Selected <> nil) and (CurrentRule <> nil);
     end;
end;

procedure TRulesEditDlg.FormShow(Sender: TObject);
begin
     actClearRules.Enabled := IE.Rules.Count > 0;
     if FirstShow and MultiEdit then
     begin
          tvRules.Selected := tvRules.Items.GetFirstNode;
          FirstShow := False;
     end;
end;

procedure TRulesEditDlg.actClearRulesExecute(Sender: TObject);
begin
     tvRules.Selected := nil;
     IE.Rules.Clear;
end;

procedure TRulesEditDlg.SetMultiEdit(const Value: Boolean);
begin
     if Value <> MultiEdit then
     begin
          FMultiEdit := Value;
          gbRules.Visible := MultiEdit;
          pnlLoadSave.Visible := MultiEdit;
          pnlTransfer.Visible := MultiEdit;
          if MultiEdit then
             bbHelp.HelpContext := 12
          else
             bbHelp.HelpContext := 14;
          if MultiEdit and FirstShow then
          begin
               tvRules.Selected := tvRules.Items.GetFirstNode;
               FirstShow := False;
          end;
     end;
end;

procedure TRulesEditDlg.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
     if (not MultiEdit) and (ModalResult = mrOk) then
        UpdateCurrentRule;
end;

procedure TRulesEditDlg.tvRulesKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
     if Key in [vk_Delete, vk_Back] then
        actDeleteRule.Execute;
end;

procedure TRulesEditDlg.lbFunctionsClick(Sender: TObject);
begin
     with Sender as TListBox do
          if ItemIndex > -1 then
             mmFunctionDesc.Lines.Text := IE.UserFunctions[Items[ItemIndex]].Comment;
end;

end.
